# frozen_string_literal: true

# Legende
# Les script a plusieurs parties indique que les bash (Batch dans l'app)
# peuvent s'initier avec differents types de source
# Le besoin est donc d'isoler la logique associee au modele actuel
# ne pas oublier d'indexer les champs simples

# TODO: Réflechir au set de housing_requests_ids (validate main_applicant)
#
#
module Hermes
  # Applicant class
  class Applicant
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::Status
    include ::Concerns::Geolocalizable
    include ::Concerns::Notification::Targetable
    include Hermes::Concerns::LockableConcern

    hermes_tools 'revenues'
    
    
    field :test_created_at, type: Date

    field :temporary_id, type: String

    field :nud, type: String
    field :sne_profile_creation_date, type: DateTime
    field :ss_id, type: String

    field :first_name, type: String
    field :last_name, type: String
    field :full_name, type: String
    field :patronymic_name, type: String
    field :nationality, type: String
    field :sms_notification, type: Boolean
    field :email, type: String
    field :email_advances_assistant, type: String
    field :event_notification, type: Boolean

    field :cellphone, type: String
    field :homephone, type: String
    field :workphone, type: String

    field :civility, type: String

    field :birth_date, type: DateTime
    field :birth_plus_30, type: DateTime

    field :work_contract_type, type: String
    field :profession_details, type: Hash
    field :has_other_employers, type: Boolean
    field :main_employer_name, type: String
    field :main_employer_siret, type: String
    field :workplace_postal_code, type: String
    field :workplace_locality, type: String
    field :foreign_workplace_postal_code, type: String
    field :foreign_workplace_locality, type: String
    field :foreign_workplace_country, type: String

    field :year_minus_one, type: String
    field :amount_year_minus_one, type: Float
    field :year_minus_two, type: String
    field :amount_year_minus_two, type: Float
    field :ressources_details, type: Hash
    field :monthly_incomes, type: Hash
    field :monthly_charges, type: Hash
    field :tax_revenues, type: Float
    field :apl_amount, type: Float
    field :family_situation, type: String
    field :family_situation_validated, type: Boolean

    field :asked_children_nb, type: Integer
    field :valid_children_nb, type: Integer
    field :asked_family_size, type: Integer
    field :valid_family_size, type: Integer

    field :situation_tax_treshold, type: Integer
    field :asked_locataire_hlm, type: Boolean
    field :valid_locataire_hlm, type: Boolean
    field :valid_priority_public, type: Boolean
    field :asked_priority_public, type: Boolean

    field :single_parent, type: Boolean
    field :priority_public, type: Boolean
    field :mobility, type: Boolean
    field :mutation, type: Boolean

    field :canceled_application_nb, type: Integer, default: 0 # Increment it when applicant cancel a housing_application
    field :completed_requests_nb, type: Integer
    field :on_going_request_nb, type: Integer

    field :actual_housing_details, type: Hash
    field :postal_details, type: Hash
    field :is_owner, type: Boolean
    field :owned_housing_details, type: Hash

    field :disability, type: String
    field :renewal_date, type: DateTime

    field :eligibility_query, type: String, default: '{"is_public":true}' # JSON formated as string

    field :main_user_id, type: BSON::ObjectId
    field :workplace_insee_code, type: String

    alias_attribute :district, :city

    common_indexes *%i[
      nud
      sne_profile_creation_date
      ss_id
      first_name
      last_name
      full_name
      patronymic_name
      nationality
      email
      event_notification
      cellphone
      civility
      birth_date
      work_contract_type
      has_other_employers
      year_minus_one
      amount_year_minus_one
      year_minus_two
      amount_year_minus_two
      ressources_details
      tax_revenues
      apl_amount
      family_situation
      situation_tax_treshold
      status
      status_id
      status_updated_at
      old_status
      priority_public
      mobility
      mutation
      completed_requests_nb
      canceled_application_nb
      on_going_request_nb
      is_owner
      main_user_id
    ]

    attr_accessor :should_update_eligibility

    # temporary fields to handle user side effects
    field :name_will_change, type: Boolean, default: false
    field :email_will_change, type: Boolean, default: false

    has_many(
      :housing_requests,
      class_name: 'Hermes::HousingRequest'
    )
    belongs_to(
      :entity,
      class_name: 'Hermes::Entity',
      index: true
    )
    has_many(
      :scores,
      class_name: 'Hermes::Score'
    )

    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id
    writer_auth_roles 'collaborateur entreprise', "collaborateur offre logement"

    # validates_presence_of :first_name, :last_name

    ELLIGIBILY_FIELDS = %w[
      entity_id
      mutation
      mobility
      priority_public
      single_parent
    ].concat((1..12).map { |i| "treshold_category_#{i}" })

    before_save :before_actions # https://dev.hermes.openstack.local/admin/families/5d1239624ac06e4ca38c1670/mapping
    after_save :after_actions

    Applicant.add_status(:single)

    before_destroy :before_destroy_actions

    def department
      postal_code[0..1] if postal_code
    end

    def before_destroy_actions
      cleanup_user
      destroy_applicant_link
      remove_from_housing_requests
      destroy_required_documents
    end

    def destroy_applicant_link
      applicant_links = Hermes::ApplicantLink.where('$or' => [{ applicant_ids: id.to_s }, { applicant_ids: id }])
      applicant_links.each do |al|
        if al.applicant_ids.length > 1

          al.applicant_ids.delete(id.to_s)
          al.applicant_ids.delete(id)
          al.save
        else
          al.destroy
        end
      end
    end
    def destroy_required_documents
      required_documents = Hermes::RequiredDocument.where(applicant__id: id)
      required_documents.each do |rd|
        rd.destroy
      end
    end

    def remove_from_housing_requests
      housing_requests = Hermes::HousingRequest.where('$or' => [{ applicant_ids: id.to_s }, { applicant_ids: id }])
      housing_requests.each do |hr|
        hr.applicant_ids.delete(id)
        hr.applicant_ids.delete(id.to_s)
        hr.with_lock(reload: false) do
          hr.save
        end
      end
    end

    def cleanup_user
      if user && (user.main_applicant_id == id)
        errors.add(:base, 'Cannot destroy my applicant')
        false
        # self.user.set(main_applicant_id: nil)
      end
    end

    def links(type: nil)
      links = Hermes::ApplicantLink.where(applicant_ids: id.to_s)
      links = links.where(type: type) if type.present?
      Hermes::Applicant.where(:id.in => (links.pluck(:applicant_ids).flatten.compact.uniq - [id.to_s]))
    end

    def link_types_for(applicant_id)
      Hermes::ApplicantLink.where(:$and => [{ applicant_ids: id.to_s }, { applicant_ids: applicant_id.to_s }]).pluck(:type).uniq
    end

    def first_active_request
      @first_active_request ||= housing_requests.find_by(is_active: true)
    end

    def coapplicants
      first_active_request&.coapplicants
    end

    def coapplicant_ids
      first_active_request&.coapplicants.map { |a| a[:id] }
    end

    def monthly_revenues
      (monthly_incomes || {}).values_fsum - (monthly_charges || {}).values_fsum
    end

    def inherit_user_datas
      # TODO: set to false after process
      return unless main_user

      self.first_name = main_user.first_name
      self.last_name = main_user.last_name
      self.full_name = main_user.full_name
      self.email = main_user.email
      first_active_request.name_has_changed = true if first_active_request
      self.name_will_change = false
      self.email_will_change = false
    end

    def update_birth_dependencies
      return if birth_date.blank?

      first_active_request.birth_date_has_changed = true if first_active_request
      self.should_update_eligibility = true
      self.birth_plus_30 = birth_date + 30.years
    end

    def update_employer_data
      if (entity_was = Hermes::Entity.find_by(id: entity_id_was))
        entity_was&.group&.set(
          user_ids: (entity_was.group.user_ids - [user_id]),
          updated_at: DateTime.current
        )
      end
      return unless entity

      self.main_employer_name = entity.title
      self.main_employer_siret = entity.siret
      entity&.group&.set(
        user_ids: entity.group.user_ids + [user_id],
        updated_at: DateTime.current
      )

      if entity.supervisor_group
        self.readers = ["g_#{entity.supervisor_group.id}"]
      end
      self.writers = ["g_#{entity.manager_group.id}"] if entity.manager_group
    end

    def update_entity_relationship
      self.entity = Hermes::Entity.find_by(siret: main_employer_siret)
    end

    def check_monthly_revenues_dependency
      # TODO: set to false after process
      cond = monthly_incomes_changed? || monthly_charges_changed?
      if first_active_request && cond
        first_active_request.monthly_revenues_will_change = true
      end
      self.should_update_eligibility = true
    end

    def check_category_dependency
      # TODO: set to false after process
      if first_active_request && family_situation_changed?
        first_active_request.category_may_change = true
      end
      self.should_update_eligibility = true
    end

    def check_eligibility_dependency
      # TODO: set to false after process
      cond = (changes.keys & ELLIGIBILY_FIELDS).size.positive?
      if first_active_request && cond
        first_active_request.should_update_eligibility = true
      end
      self.should_update_eligibility = true
    end

    def check_year_revenues
      cond = amount_year_minus_two_changed?
      if first_active_request && cond
        first_active_request.tax_revenues_will_change = true
      end
      self.should_update_eligibility = true
    end

    def check_all
      check_monthly_revenues_dependency
      check_category_dependency
      check_eligibility_dependency
      check_year_revenues
    end

    # def should_touch_active_request
    #   return unless first_active_request

    #   first_active_request.monthly_revenues_will_change ||
    #     first_active_request.category_may_change ||
    #     first_active_request.should_update_eligibility ||
    #     first_active_request.name_has_changed ||
    #     first_active_request.tax_revenues_will_change
    # end

    def main_user
      User.find_by(id: main_user_id)
    end

    def identity_values_changed?
      first_name_changed? || last_name_changed? || email_changed?
    end

    def user_data_moved?
      (main_user_id_changed? || name_will_change || email_will_change)
    end

    def needs_first_identity_set?
      new_record? && !identity_values_changed?
    end

    def update_fullname
      self.full_name = [first_name, last_name].join(' ')
    end

    def update_status_dependencies
      self.status_updated_at = DateTime.current
      self.old_status = status_was
    end

    def before_actions
      inherit_user_datas if user_data_moved?
      if !main_user && (first_name_changed? || last_name_changed?)
        update_fullname
      end
      update_entity_relationship if main_employer_siret_changed?
      update_employer_data if entity_id_changed?
      update_birth_dependencies if birth_date_changed?
      update_right if user_id_changed?
      check_all
    end

    def update_first_active_request_acl
      return unless first_active_request

      first_active_request.set(readers: readers, writers: writers, updated_at: DateTime.current)
    end

    def update_right
      readers.push(user_id.to_s)
    end

    def update_user_main_applicant_id
      now = DateTime.current
      anticipated_writers = writers
      if main_user_id && targeted_user = User.find_by(id: main_user_id)
        targeted_user.set(
          main_applicant_id: id,
          updated_at: now
        )
        anticipated_writers << main_user_id.to_s
      end

      if main_user_id_was && (previous_user = User.find_by(id: main_user_id_was))
        previous_user.set(
          main_applicant_id: nil,
          updated_at: now
        )
        anticipated_writers.delete(main_user_id_was.to_s)
      end

      set(writers: anticipated_writers, updated_at: now)
    end

    def update_housing_request_folders_if_needed
      is_needed = (nationality_changed? ||
        work_contract_type_changed? ||
        monthly_incomes_changed? ||
        family_situation_changed? ||
        actual_housing_details_changed? ||
        amount_year_minus_one_changed? ||
        amount_year_minus_two_changed?

                  )
      return unless is_needed

      hrs = HousingRequest.any_of({ main_applicant_id: id }, { main_applicant_id: id.to_s }, { applicant_ids: id }, applicant_ids: id.to_s)

      hrs.each do |concerned_hr|
        concerned_hr.housing_request_folder.compute_required_documents
        concerned_hr.housing_request_folder.save
      end
    end

    def prepare_first_active_request_data_changes
      if first_name_changed? ||
         last_name_changed? ||
         email_changed?
        first_active_request&.name_has_changed = true
      end

      first_active_request&.birth_date_has_changed = true if birth_date_changed?
    end

    def update_required_documents_user_id
      required_documents = Hermes::RequiredDocument.where(applicant_id: id)
      required_documents.each do |rd|
        rd.writers.push(user_id)
        rd.save
      end
    end

    def after_actions
      check_status_changed
      update_user_main_applicant_id if main_user_id_changed?
      update_first_active_request_acl if readers_changed? || writers_changed?
      prepare_first_active_request_data_changes
      if first_active_request&.changed? || should_update_eligibility
    
        first_active_request&.with_lock(reload: false) do
          first_active_request&.save
        end
      end
      set(should_update_eligibility: nil)
      update_housing_request_folders_if_needed
      update_required_documents_user_id if user_id_changed?
      true
    end
  end
end
